/****************************************************************************
PROGRAMA:    sets.c
AUTOR:       Patricio Garca Bez
FECHA:       Curso 92/93
FINALIDAD:   Librera de funciones para trabajar con conjuntos.
COMENTARIOS: Los conjuntos se implementan como punteros a una estructura
             llamada set (ver sets.h) que contiene

HISTORIA:  F. de Sande. Adaptada y documentada 28.09.94
           F. de Sande. Revisin de algn Bug 18.11.94
           F. de Sande. 04.12.96 Cambio el nombre del fichero CNJLIB.C -> SETS.C
           F. de Sande. 22.10.98 Cambio de los nombres de las rutinas cnj* -> set_*
     Juan Ramn Glez Glez. 29.10.98 Adicin de la funcin *set_int
     Juan Ramn Glez Glez. 31.10.98 Adicin de la funcin set_incl
           F. de Sande. 09.11.01 Algunos cambios superficiales. Ms documentacin. 
                                 chk_malloc
BIBLIOGRAFIA:
MODO DE UTILIZACION:
****************************************************************************/
#include <stdlib.h>
#include "basic.h"
#include "sets.h"

/****************************************************************************
* *set_ini --   Aloja memoria para un conjunto con capacidad mxima
*    de max_ele elementos
*
* Parmetros:
*    max_ele: Nmero mximo de elementos que va a almacenarse en el conjunto.
* Devuelve:
*    Un puntero a una estructura de tipo conjunto que se ha inicializado
*    convenientemente.
****************************************************************************/
set *set_ini(unsigned max_ele) {
  palabra *p;
  set *cj;
  unsigned l;

  l = PAL(max_ele - 1) + 1;                         /* No. de palabras necesarias */
  p = (palabra *) chk_malloc(l * sizeof(palabra));

  cj = (set *) chk_malloc(sizeof(set));
  cj->longitud = l;
  cj->conj = p;
  return (cj);
}

/****************************************************************************
* set_per -- Indica != 0  0 segun el elemento nmero e pertenezca
    *        o no en el conjunto cj
*
* La pertenencia se da si el bit correspondiente est a 1, lo cual se
* comprueba haciendo un AND de bits con 1 rotado a la izquierda el nmero de
* bits necesario con la palabra correspondiente al elemento e.
*
* Parmetros:
*      *cj -- Puntero al conjunto considerado
*      e   -- Elemento cuya pertenencia se estudia
*****************************************************************************/
boolean set_per(set *cj, unsigned e) {
  return((boolean)(cj->conj[PAL(e)] & (1UL << BIT(e))) != 0);
}
/****************************************************************************
* set_ins -- Aade el elemento nmero e al conjunto cj
*
* Para ello se hace un OR de bits en la palabra y bits adecuados. Si el
* elemento ya perteneca al conjunto, esta operacin no provocar ningn
* cambio.
*
* Parmetros:
*      *cj -- Puntero al conjunto considerado
*      e   -- Elemento a aadir
*****************************************************************************/
void set_ins(set *cj, unsigned e) {
  unsigned i;
  palabra j;

  i = PAL(e);
  j = (1UL << BIT(e));
  cj->conj[i] |= j;
}
/****************************************************************************
* set_vac -- Inicializa a vaco el conjunto
*
* La operacin se implementa poniendo a 0 todas las palabras del conjunto.
*
* Parmetros:
*      *cj -- Puntero al conjunto considerado
*****************************************************************************/
void set_vac(set *cj) {
  palabra *i;
  unsigned j;

  j = cj->longitud;
  i = cj->conj;
  while (j--)
    *i++ = 0;
}
/****************************************************************************
* set_cpy -- Copia el conjunto cjorg en el conjunto cjdest
*
* En la implementacin se est suponiendo implcitamente que el conjunto
* destino tiene una longitud mayor o igual que la del origen.
*
* Parmetros:
*      *cjdest -- Puntero al conjunto destino
*      *cjorg -- Puntero al conjunto origen
*****************************************************************************/
void set_cpy(set *cjorg, set *cjdest) {
  unsigned k;
  palabra *i, *j;

  k = cjorg->longitud;
  j = cjorg->conj;
  i = cjdest->conj;
  while (k--)
    *i++ = *j++;
  k = (cjdest->longitud) - (cjorg->longitud);
  while (k--)
    *i++ = 0;
}
/****************************************************************************
* set_igu -- Devuelve <>0  0 segn sean iguales o no los conjuntos cj1 y cj2
*
* En primer lugar se devuelve 0 si el nmero de palabras de los conjuntos es
* distinto. Si son iguales se devuelve 0 en cuanto haya dos palabras distintas
* o bien 1 si todas son iguales.
*
* Parmetros:
*      *cj1 -- Puntero al primer conjunto
*      *cj2 -- Puntero al segundo conjunto
*****************************************************************************/
boolean set_igu(set *cj1, set *cj2) {
  unsigned k;
  palabra *i, *j;

  if ((k = cj1->longitud) != (cj2->longitud))
    return(0);
  j = cj1->conj;
  i = cj2->conj;
  while ((--k) && (*i == *j)) {
    i++;
    j++;
  }
  return ((boolean)(*i == *j));
}
/*****************************************************************************
* set_lib -- Libera la memoria reservada a un conjunto mediante set_ini
*
* Parmetros:
*      *cj -- Puntero al conjunto
*****************************************************************************/
void set_lib(set *cj) {
  free(cj->conj);
  free(cj);
}
/****************************************************************************
* set_uni -- Aade los elelmentos del conjunto cjorg en el conjunto cjdest
*
* La unin se implementa como un OR de bits de las palabras de ambos conjuntos
*
* Parmetros:
*      *cjdest -- Puntero al conjunto destino
*      *cjorg -- Puntero al conjunto origen
*****************************************************************************/
void set_uni( set *cjdest, set *cjorg) {
  unsigned k;
  palabra *i, *j;

  k = cjorg->longitud;
  j = cjorg->conj;
  i = cjdest->conj;
  while (k--)
    *i++ |= *j++;
}
/****************************************************************************
* set_print -- Imprime en pantalla los elementos del conjunto
*
* Parmetros:
*      *s -- Puntero al conjunto a imprimir
*****************************************************************************/
void set_print(set *s) {
  int i,
    n_bits;   /* Nmero total de bits en la representacin del conjunto */

  n_bits = (s->longitud) * sizeof(palabra) * BITS_EN_BYTE;
  printf ("{");
  for (i = 0; i < n_bits; i++) {
   if ((s->conj[PAL(i)] & (1UL << BIT(i))) != 0)   /* Si el elemento est en el cj */
    printf (" %d ", i);
  }
  printf ("}");
}
/****************************************************************************
* *set_int -- Halla la interseccin de dos conjuntos.
*
* La interseccin se implementa como un AND de bits de las palabras de ambos
* conjuntos.
*
* Parmetros:
*      *cj1 -- Puntero al conjunto 1
*      *cj2 -- Puntero al conjunto 2
*
* Devuelve:
*        Un puntero al conjunto resultado de la interseccin.
*****************************************************************************/
set *set_int(set *cj1, set *cj2) {
  set *cjfinal; /* Conjunto interseccin de cj1 y cj2 */
  unsigned long_mayor, long_menor; /* Longitud del mayor y del menor de
                          los dos conjuntos.          */
  palabra *pal_cj1, *pal_cj2, *pal_fin;
    /* Punteros a la palabra siendo intersectada en cada conjunto. */

  if (cj1->longitud >= cj2->longitud) {
    long_mayor = cj1->longitud;
    long_menor = cj2->longitud;
  }
  else {
    long_mayor = cj2->longitud;
    long_menor = cj1->longitud;
  }
  if ((cjfinal = set_ini ((long_mayor * sizeof(palabra) * BITS_EN_BYTE))) == NULL) {
    /* El conjunto final es del mismo tamao que el mayor */
    printf ("Insuficiente memoria.\n");
    exit (0);
  }
  set_vac (cjfinal); /* Se vaca el conjunto final porque si los conjun-
                 tos no son del mismo tamao las ltimas palabras
                 estaran indeterminadas.                        */
  pal_cj1 = cj1->conj;
  pal_cj2 = cj2->conj;
  pal_fin = cjfinal->conj;
  for (; long_menor > 0; long_menor--)
    /* Slo es necesario hacer la interseccin del nmero de palabras
      del menor. */
    *pal_fin++ = *pal_cj1++ & *pal_cj2++;
  return (cjfinal);
}
/****************************************************************************
* set_incl -- Indica si el conjunto menor est incluido en el conjunto mayor
*
* La inclusin se comprueba interceptando cjmenor con cjmayor y comprobando
* que el resultado sigue conteniendo todos los elementos de cjmenor.
*
* Parmetros:
*      *cjmenor -- Conjunto a mirar si est incluido
*      *cjmayor -- Conjunto que debe contener al conjunto menor
*
* Devuelve:
*        1 si cjmenor est incluido en cjmayor  0 si no lo est.
*****************************************************************************/
boolean set_incl (set *cjmenor, set *cjmayor) {
  set *cjaux;
  boolean val_devuelto;

  cjaux = set_int (cjmenor, cjmayor);
  cjaux->longitud = cjmenor->longitud;
    /* Se iguala la longitud del conjunto auxiliar a la de cjmenor para
      que el hecho de que ste conjunto tenga ms palabras no afecte a
      la comparacin que sigue.                           */
  if (set_igu (cjaux, cjmenor))
    val_devuelto = TRUE;
  else
    val_devuelto = FALSE;
  set_lib (cjaux);
  return val_devuelto;
}
/****************************************************************************
* *set_most -- Muestra el conjunto set_amost, mostrando todos sus elementos
*    encerrados entre llaves. Muestra el conjunto en la posicin
*    actual de la pantalla, sin introducir retornos de carro o es-
*    pacios ni antes ni despus del conjunto.
*
* Parmetros:
*    *set_amost -- Puntero al conjunto que se quiere mostrar
*****************************************************************************/
void set_most (set *set_amost) {
  int i;

  printf ("{");
  for (i = 0; i < (set_amost->longitud * sizeof(palabra) * BITS_EN_BYTE); i++) {
    if (set_per (set_amost, i))
      printf (" %d ", i);
  }
   printf ("}");
}
/****************************************************************************
* set_is_vac -- Devuelve 1 si el conjunto cj est vaco  0 si no lo est.
*
* Parmetros:
*        *cj -- Puntero al conjunto considerado
*****************************************************************************/
boolean set_is_vac(set *cj) {
  palabra *pal;                   /* Puntero a la palabra considerada */
  unsigned i;
  boolean EsVacio = TRUE;

  for (i = 0, pal = cj->conj; (i < cj->longitud) && EsVacio; i++, pal++) {
    if (*pal != 0) {
      EsVacio = FALSE;
      break;
    }
  }
  return (EsVacio);
}
